from collections import defaultdict
from itertools import product

import numpy as np
import pandas as pd
from sympy  import sympify, lambdify, Symbol, Piecewise

from phi_relative_lcmsms.helper import create_pivot

def calc_identifier_with_glucose_tca(nacorr_merged_output_df = None, metadata_mq_df = None, 
                                     phi_expressions_df = None, interm_expressions_df = None):
    """
    This function will calculate the identifier expressions using quantitative data for glucose labeled data
    :param nacorr_merged_output_df: The na correction output
    :param metadata_mq_df: The metadata of isotopologues/fragments with their unique identifiers
    :param phi_expressions_df: A dataframe having identifier expressions for phis
    :param interm_expressions_df: A dataframe having identifier expressions for intermediates
    :return: A list of dataframes for calculated identifier from expressions and missing identifers
    """

    print ("Performing Preprocessing...\n")
    fe_sample_identifier_dict = make_fe_sample_identifier_dict(nacorr_merged_output_df, metadata_mq_df)
    samples_list = list(fe_sample_identifier_dict.keys())

    fe_sample_identifier_temp_dict = fe_sample_identifier_dict.copy()
    
    for sample in samples_list[0:1]:
        fe_sample_identifier_temp_dict[sample]['Phi_AcCit'] = 'Phi_AcCit'    
    validated_interm_exp_list = validate_identifier_variables(interm_expressions_df, fe_sample_identifier_temp_dict)
    filtered_interm_expressions_df = validated_interm_exp_list[0]
    missing_interm_expressions_df = validated_interm_exp_list[1]
    
    print ("Citrate Deconvolution Started...\n")
    cit_deconv_dict = citrate_deconvolution(filtered_interm_expressions_df, fe_sample_identifier_dict)
    print ("Citrate Deconvolution Completed...\n")
    samples_list = list(fe_sample_identifier_dict.keys())
    fe_cit_deconv_merge_dict = defaultdict(dict)
    for sample in samples_list:
        fe_cit_deconv_merge_dict[sample] = {**fe_sample_identifier_dict[sample], **cit_deconv_dict[sample]}
    print ("Phi Calculation Started...\n")
    
    validated_phi_exp_list = validate_identifier_variables(phi_expressions_df, fe_cit_deconv_merge_dict)
    filtered_phi_expressions_df = validated_phi_exp_list[0]
    missing_phi_expressions_df = validated_phi_exp_list[1]
    
    calculated_phi_dict = calculate_identifier_from_expression(filtered_phi_expressions_df, fe_cit_deconv_merge_dict)
    print ("Phi Calculation Completed...\n")
    
    interm_identifier_list = ['Phi_AcCit'] + filtered_interm_expressions_df['Identifier'].tolist()
    cit_deconv_interm_df = pd.DataFrame(cit_deconv_dict)
    cit_deconv_interm_df = cit_deconv_interm_df.reindex(interm_identifier_list)
    cit_deconv_interm_df = cit_deconv_interm_df.rename_axis('Identifier').reset_index()
    
    phi_identifier_list = filtered_phi_expressions_df['Identifier'].tolist()
    calculated_phi_df = pd.DataFrame(calculated_phi_dict)
    calculated_phi_df = calculated_phi_df.reindex(phi_identifier_list)
    calculated_phi_df = calculated_phi_df.rename_axis('Identifier').reset_index()
    
    calculated_identifier_df = pd.concat([calculated_phi_df, cit_deconv_interm_df], ignore_index = True)
    
    missing_identifier_df = pd.concat([missing_interm_expressions_df, missing_phi_expressions_df], ignore_index = True)
    
    return [calculated_identifier_df, missing_identifier_df]

def calc_identifier_generic(nacorr_merged_output_df = None, metadata_mq_df = None, 
                             identifier_expressions_df = None):
    """
    This function will calculate the identifier expressions using quantitative data for generic labeled data
    :param nacorr_merged_output_df: The na correction output
    :param metadata_mq_df: The metadata of isotopologues/fragments with their unique identifiers
    :param phi_expressions_df: A dataframe having identifier expressions for phis
    :param interm_expressions_df: A dataframe having identifier expressions for intermediates
    :return: A list of dataframes for calculated identifier from expressions and missing identifers
    """
    
    print ("Identifier Calculation Started...\n")

    fe_sample_identifier_dict = make_fe_sample_identifier_dict(nacorr_merged_output_df, metadata_mq_df)
    validated_identifier_exp_list = validate_identifier_variables(identifier_expressions_df, fe_sample_identifier_dict)
    filtered_identifier_expressions_df = validated_identifier_exp_list[0]
    missing_identifier_df = validated_identifier_exp_list[1]
    
    calculated_identifier_dict = calculate_identifier_from_expression(filtered_identifier_expressions_df, fe_sample_identifier_dict)

    filtered_identifier_list = filtered_identifier_expressions_df['Identifier'].tolist()
    calculated_identifier_df = pd.DataFrame(calculated_identifier_dict)
    calculated_identifier_df = calculated_identifier_df.reindex(filtered_identifier_list)
    calculated_identifier_df = calculated_identifier_df.rename_axis('Identifier').reset_index()
    
    print ("Identifier Calculation Completed...\n")
    
    return [calculated_identifier_df, missing_identifier_df]

def calculate_identifier_from_expression(identifier_expressions_df, fe_sample_identifier_dict):
    """
    This function will calculate the identifier expressions using quantitative data
    :param identifier_expressions_df: A dataframe having identifier expressions
    :param fe_sample_identifier_dict: A dictionary with samples as keys and isotopoloues/fragments with their quantitative values
    :return: A dictionary of calculated identifier from expression
    """

    identifier_expressions_calculated = pd.DataFrame()
    all_identifier_exp_calc_dict = fe_sample_identifier_dict.copy()
    identifier_expressions_calculated_dict = defaultdict(dict)
    identifier_list = identifier_expressions_df['Identifier'].unique().tolist()
    samples_list = list(fe_sample_identifier_dict.keys())
    
    for sample, identifier in product(samples_list, identifier_list):
        individual_identifier_expression_df = identifier_expressions_df[identifier_expressions_df['Identifier']==identifier]
        expression_str =individual_identifier_expression_df['Expression'].values[0]
        variables_string = individual_identifier_expression_df['Variables_Name'].values[0]
        variables_name_dict = create_variables_from_string(variables_string)
        variables_name_list = list(variables_name_dict.keys())
        subset_fe_sample_dict = dict((k, all_identifier_exp_calc_dict[sample][k]) for k in variables_name_list)
        vairables_keys = list(subset_fe_sample_dict.keys())
        variables_values = np.array(list(subset_fe_sample_dict.values()))
        lambdify_input_exp = get_lambdify_input_expression(expression_str)
        cal_exp_funtion = lambdify(vairables_keys, lambdify_input_exp, 'numpy')
        sample_identifier_value = float(cal_exp_funtion(*variables_values))
        all_identifier_exp_calc_dict[sample][identifier] = sample_identifier_value
        identifier_expressions_calculated_dict[sample][identifier] = sample_identifier_value
    
    return (identifier_expressions_calculated_dict)

def citrate_deconvolution(identifier_expressions_df, fe_sample_identifier_dict):
    """
    This function will calculate the identifier expressions for intermediates using quantitative data
    :param identifier_expressions_df: A dataframe having identifier expressions
    :param fe_sample_identifier_dict: A dictionary with samples as keys and isotopoloues/fragments with their quantitative values
    :return: A dictionary of calculated identifier from expression for intermediates
    """

    samples_list = list(fe_sample_identifier_dict.keys())
    print ("Total number of samples :", len(samples_list), "\n")
    
    cit_deconv_dict = defaultdict(dict)
    
    def iterate_phi_accit(phi_AcCit_val_list, subset_sample_identifier_exp_dict, sample):
        phi_AcCit_cmp_val = 0
        phi_AcCit_cmp_calc = 0
        phi_AcCit_cmp_diff = 100000
        subset_fe_sample_ident_dict = subset_sample_identifier_exp_dict.copy()
        for phi_AcCit in phi_AcCit_val_list:
            phi_AcCit = abs(phi_AcCit)
            subset_fe_sample_ident_dict[sample]['Phi_AcCit'] = phi_AcCit
            calc_expressions_dict = calculate_identifier_from_expression(identifier_expressions_df, subset_fe_sample_ident_dict)    
            phi_AcCit_cal = calc_expressions_dict[sample]['Phi_AcCit_cal']
            if not np.isnan(phi_AcCit_cal):
                phi_AcCit_diff = abs(phi_AcCit - phi_AcCit_cal)
                if phi_AcCit_cmp_diff >= phi_AcCit_diff:
                    phi_AcCit_cmp_diff = phi_AcCit_diff
                    phi_AcCit_cmp_val = phi_AcCit
                    phi_AcCit_cmp_calc = phi_AcCit_cal
        print (phi_AcCit_cmp_val, phi_AcCit_cmp_calc)
        return [phi_AcCit_cmp_val, phi_AcCit_cmp_calc]
    
    count = 1
    for sample in samples_list:
        print (count," Starting citrate deconvolution of sample :", sample, "\n")
        count = count + 1
        subset_sample_identifier_exp_dict = {sample: fe_sample_identifier_dict[sample]}
        phi_AcCit_val_0_1_list = np.round(np.arange(1,0,-0.01), 2)
        phi_AcCit_iterate_val_list = iterate_phi_accit(phi_AcCit_val_0_1_list, subset_sample_identifier_exp_dict, sample)
        phi_AcCit_cmp_0_1_val = phi_AcCit_iterate_val_list[0]
        phi_AcCit_2nd_min = phi_AcCit_cmp_0_1_val - 0.01
        if phi_AcCit_2nd_min < 0: phi_AcCit_2nd_min = 0
        phi_AcCit_2nd_max = phi_AcCit_cmp_0_1_val + 0.01
        phi_AcCit_2nd_itr_range = np.round(np.arange(phi_AcCit_2nd_min, phi_AcCit_2nd_max, 0.001), 3)
        phi_AcCit_2nd_itr_val_list = iterate_phi_accit(phi_AcCit_2nd_itr_range, subset_sample_identifier_exp_dict, sample)
        phi_AcCit_cmp_val = phi_AcCit_2nd_itr_val_list[0]
        phi_AcCit_cmp_calc = phi_AcCit_2nd_itr_val_list[1]
        subset_sample_identifier_exp_dict[sample]['Phi_AcCit'] = phi_AcCit_cmp_val
        cit_deconv_sample_dict = calculate_identifier_from_expression(identifier_expressions_df, subset_sample_identifier_exp_dict)
        cit_deconv_sample_dict[sample]['Phi_AcCit'] = phi_AcCit_cmp_val
        cit_deconv_dict[sample] = cit_deconv_sample_dict[sample]
        print ("Predicted Phi_AcCit :", phi_AcCit_cmp_val)
        print ("Calculated Phi_AcCit :", phi_AcCit_cmp_calc)
        print ("\n")
    return (cit_deconv_dict)

def create_variables_from_string(string_s):
    """
    This function will create a dictionary of variables from string separated by comma.
    :param string_s: The string having variables separated by comma
    :return: A dictionary of variables from string separated by comma
    """

    all_variables = {}
    str_s_split = string_s.split(",")
    for var_str in str_s_split:
        var_name = var_str.strip()
        latex_string = var_name.replace("_", "\_")
        all_variables[var_name] = Symbol(latex_string)

    return (all_variables)

def get_lambdify_input_expression(string_s, variables_dict = None):
    """
    This function will create a lambdify expression from string expression.
    :param string_s: The string expression
    :param variables_dict: A dictionary with defined variables
    :return: A lambdify expression from string expression
    """

    if '|' in string_s:
        expression_with_condition_list = []
        str_s_split = string_s.split("|")
        for var_str in str_s_split:
            var_name = var_str.strip()
            expression_with_condition_list.append(var_name)
        if variables_dict:
            sympify_exp_list = sympify(expression_with_condition_list, locals = variables_dict)
        else:
            sympify_exp_list = sympify(expression_with_condition_list)
        piecewise_exp_list = Piecewise(*sympify_exp_list)
        
        return (piecewise_exp_list)
    else:
        if variables_dict: 
            sympify_exp = sympify(string_s, locals = variables_dict)
        else: 
            sympify_exp = sympify(string_s)
            
        return (sympify_exp)


def make_fe_sample_identifier_dict(flux_fe_ident_df, metadata_mq_ident_df):
    """
    This function will create a dictionary with samples as keys and isotopoloues/fragments with their fractional enrichemt.
    :param flux_fe_ident_df: The na correction output
    :param metadata_mq_ident_df: The metadata of isotopologues/fragments with their unique identifiers
    :return: A dictionary with samples as keys and isotopoloues/fragments with their fractional enrichemt
    """

    flux_fe_ident_id_df = flux_fe_ident_df.copy()
    metadata_mq_ident_id_df = metadata_mq_ident_df.copy()
    all_fe_sample_dict = dict()

    metadata_mq_ident_id_df[['Metab', 'Component Name']] = metadata_mq_ident_id_df[['Metab', 'Component Name']].apply(lambda x: x.str.upper())
    flux_fe_ident_id_df[['Metab', 'Component Name']] = flux_fe_ident_id_df[['Metab', 'Component Name']].apply(lambda x: x.str.upper())
    flux_fe_ident_id_df = pd.merge(flux_fe_ident_id_df, metadata_mq_ident_id_df, on=['Metab', 'Component Name'])
    
    isotopologue_sample_df = create_pivot(flux_fe_ident_id_df, "Fractional Enrichment with zero", "Isotopologue_Id", "Sample", np.sum)
    component_name_sample_df = create_pivot(flux_fe_ident_id_df, "Fractional Enrichment with zero", "Component_Name_Id", "Sample", np.mean)

    samples_list = flux_fe_ident_id_df['Sample'].unique().tolist()
    
    for sample in samples_list:
        isotopologue_sample_dict = dict(zip(isotopologue_sample_df["Isotopologue_Id"], isotopologue_sample_df[sample]))
        component_name_sample_dict = dict(zip(component_name_sample_df["Component_Name_Id"], component_name_sample_df[sample]))
        merge_dict = {**isotopologue_sample_dict, **component_name_sample_dict}
        all_fe_sample_dict[sample] = merge_dict

    return (all_fe_sample_dict)

def sympify_exp_from_expression(identifier_expressions_df, identifier):
    """
    This function will create a sympify expression from string expression.
    :param identifier_expressions_df: A dataframe having identifier expressions
    :param identifier: An identifier for the expression
    :return: A sympify expression from string expression
    """

    individual_identifier_expression_df = identifier_expressions_df[identifier_expressions_df['Identifier']==identifier]
    expression_str =individual_identifier_expression_df['Expression'].values[0]
    variables_string = individual_identifier_expression_df['Variables_Name'].values[0]
    variables_name_dict = create_variables_from_string(variables_string)
    sympify_expression_str = get_lambdify_input_expression(expression_str, variables_dict = variables_name_dict)
    
    return (sympify_expression_str)

def validate_identifier_variables(identifier_expressions_df, fe_sample_ident_dict):
    """
    This function will varify whether the identifier expressions are valide or not.
    :param identifier_expressions_df: A dataframe having identifier expressions
    :param fe_sample_identifier_dict: A dictionary with samples as keys and isotopoloues/fragments with their quantitative values
    :return: A list of dataframes with valid and invalid identifiers
    """

    identifier_list = identifier_expressions_df['Identifier'].unique().tolist()
    samples_list = list(fe_sample_ident_dict.keys())
    filtered_identifier_exp_calc_dict = fe_sample_ident_dict[samples_list[0]].copy()
    filtered_identifier_expressions_df = identifier_expressions_df.copy()
    missing_var_df = pd.DataFrame()
    identifier_with_var_list = []
    for identifier in identifier_list:
        individual_identifier_expression_df = identifier_expressions_df[identifier_expressions_df['Identifier']==identifier]
        expression_str =individual_identifier_expression_df['Expression'].values[0]
        variables_string = individual_identifier_expression_df['Variables_Name'].values[0]
        variables_name_list = create_variables_from_string(variables_string)
        missing_var_list = [var for var in variables_name_list if var not in filtered_identifier_exp_calc_dict.keys()]
        if missing_var_list:
            interm_df = pd.DataFrame({'Identifier': identifier, 'Variables_Name': [", ".join(missing_var_list)]})
            missing_var_df = pd.concat([missing_var_df, interm_df], ignore_index = True)
        else :
            identifier_with_var_list.append(identifier)
            filtered_identifier_exp_calc_dict[identifier] = identifier
    filtered_identifier_expressions_df = filtered_identifier_expressions_df[filtered_identifier_expressions_df['Identifier'].isin(identifier_with_var_list)]
    
    return [filtered_identifier_expressions_df, missing_var_df]
